/*
 * wysmod - jiffies.c
 * 	version 0.0.0
 * 
 * Copyright (C) 2005, 2006, 2007  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <tmos/config.h>
#include <tmos/jiffies.h>
#include <tmos/serial.h>
#include <asm/system.h>
#include <asm/8051.h>
#include <tmos/sched.h>

#define __RELOAD_VALUE (CONFIG_OSCILLATOR / CONFIG_TIMER_DIVISOR / HZ)
#define RELOAD_VALUE ((uint16_t)(65536 - (__RELOAD_VALUE)))

struct __check_reload_value {
	char a[(__RELOAD_VALUE < 65536) ? 1 : -1];
};

jiffies_t jiffies = (jiffies_t)(-30 * HZ);
#if CONFIG_JIFFIES_SEQLOCK
uint8_t jiff_seq = 0;

/*
 * FIXME: before jiff_seq++ should be smp_wmb(), but current barrier()
 * implementation causes storing too many registers.
 *
 * Current versions of SDCC compiles good code without smp_wmb().
 * jiff_seq must be incremented AFTER incrementing jiffies. This
 * important not only on SMP (currently not supported), but also
 * when timer interrupt is normal interrupt and high priority
 * interrupt occur.
 */
#define jiff_seq_inc()	do { jiff_seq++; } while (0)

#else
#define jiff_seq_inc()
#endif

extern struct context __kernel *current;

#if CONFIG_USER_PREEMPT && !CONFIG_KERNEL_PREEMPT
#define timer_irq_preempt_enabled() \
	preempt_enabled_flags(*((unsigned char __data *) SP))
#else
#define timer_irq_preempt_enabled() preempt_enabled()
#endif

void timer_irq(void) __interrupt(1) __using(TIMER_BANK)
{
	TL0 = RELOAD_VALUE & 0xff;
	TH0 = RELOAD_VALUE >> 8;

	jiffies++;
	jiff_seq_inc();

#ifdef CONFIG_TASK_TIMES
	current->cpu_time++;
#endif

	/* FIXME: add generic timers */
#ifdef CONFIG_TRT_BIMP_VELOCITY
	trt_bimp_irq();
#endif

	irq_switch_task();
	set_need_wakeup();
	TF0 = 0;
}

#if CONFIG_BITS_PER_JIFFIES != 8
#if CONFIG_JIFFIES_SEQLOCK

jiffies_t get_jiffies(void) REENTRANT
{
	uint8_t seq;
	jiffies_t ret;

	do {
		seq = jiff_seq;
		smp_rmb();
		ret = jiffies;
		smp_rmb();
	} while (seq != jiff_seq);
	return ret;
}

#elif CONFIG_JIFFIES_TIMER_IRQ

jiffies_t get_jiffies(void) REENTRANT
{
	jiffies_t ret;

	preempt_disable();
	ET0 = 0;
	ret = jiffies;
	ET0 = 1;
	preempt_enable();
	return ret;
}

#elif CONFIG_JIFFIES_LOCAL_IRQ

jiffies_t get_jiffies(void) REENTRANT
{
	jiffies_t ret;
	cpu_flags_t flags;

	local_irq_save(flags);
	ret = jiffies;
	local_irq_restore(flags);
	return ret;
}

#else
#error "Unsupported jiffies synchronization method"
#endif
#endif

void jiffies_init(void)
{
	TR0 = 0;
	TMOD = (TMOD & ~0x0f) | 0x01;
	TL0 = RELOAD_VALUE & 0xff;
	TH0 = RELOAD_VALUE >> 8;
	TR0 = 1;
	TF0 = 1;
	ET0 = 1;
#if CONFIG_TIMER_HIGHPRIO
	PT0 = 1;
#endif
}

void jsleep(jiffies_t t) REENTRANT
{
	t += get_jiffies();
	do {
		schedule();
	} while (!time_after_eq(get_jiffies(), t));
}

u16 tsc_read(void)
{
	u8 low;
	u8 high;

	do {
		high = TH0;
		low = TL0;
	} while (high != TH0);
	return pack16(low, high);
}

u16 tsc_read_low(void) __using(LOW_PRIORITY_BANK)
{
	u8 low;
	u8 high;

	do {
		high = TH0;
		low = TL0;
	} while (high != TH0);
	return pack16(low, high);
}
