/*
 * tmos - port.h 
 * Copyright (C) 2009  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Fundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef __TMOS_PORT_H
#define __TMOS_PORT_H

#include <tmos/kernel.h>
#include <tmos/types.h>
#include <tmos/jiffies.h>

#ifdef CONFIG_PORT
typedef signed char msg_ssize_t;
typedef unsigned char msg_size_t;
typedef unsigned char msg_flags_t;

#define O_RDONLY		0
#define O_WRONLY		0
#define O_RDWR			0

#if CONFIG_OFFSET_32
typedef int32_t off_t;
#define OFF_MAX INT32_MAX
#elif CONFIG_OFFSET_16
typedef int16_t off_t;
#define OFF_MAX INT16_MAX
#elif CONFIG_OFFSET_8
typedef int8_t off_t;
#define OFF_MAX INT8_MAX
#else
#error "unsupported offset size"
#endif

#define MSG_DONTWAIT		1

typedef signed char port_t;

typedef msg_ssize_t (*port_xmit_t)(port_t port, char __kernel *buf,
		msg_size_t len, msg_size_t max_resp_len, msg_flags_t flags,
		off_t off)
	__reentrant;
extern const port_xmit_t port_xmit_table[];

port_t __port_open(const char *__kernel name, int flags) __reentrant;
#define port_open(name, flags, ...) __port_open(name, flags)

port_t __port_open_c(const char *__kcode name, int flags) __reentrant;
#define port_open_c(name, flags, ...) __port_open(name, flags)

#if CONFIG_OPTIMIZE_SIZE
msg_ssize_t port_pxmit(port_t port, char __kernel *buf, msg_size_t len,
		msg_size_t max_resp_len, msg_flags_t flags, off_t off)
	__reentrant;
#else
#define port_pxmit(port, buf, len, max_resp_len, flags, off) \
	port_xmit_table[port](port, buf, len, max_resp_len, flags, off)
#endif

#define port_xmit(port, buf, len, max_resp_len, flags) \
	port_pxmit(port, buf, len, max_resp_len, flags, 0)

#define port_write(port, buf, len) \
	port_xmit((port), (buf), (len), 0, 0)
#define port_read(port, buf, len) \
	port_xmit((port), (buf), 0, (len), 0)
#define port_pwrite(port, buf, len, off) \
	port_pxmit((port), (buf), (len), 0, 0, off)
#define port_pread(port, buf, len, off) \
	port_pxmit((port), (buf), 0, (len), 0, off)

msg_ssize_t port_read_t(port_t port, char __kernel *buf, msg_size_t len,
		jiffies_t timeout) __reentrant;

#define PORT_2(path, xmit) \
	msg_ssize_t xmit(port_t, char __kernel *, msg_size_t, msg_size_t, \
			msg_flags_t, off_t) __reentrant;
#include <tmos/ports.h>
#undef PORT_2
#endif

#endif
